#include "cora_fit.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <float.h>
#include <unistd.h>

#include <math.h>

float clamp_float (float minimum, float x, float maximum, char *pname)
{
  if (x > maximum || x < minimum)
    {
      if (maximum < FLT_MAX)
        fprintf(stderr, "*  Warning  *: %s:  %f: out of range [%f...%f]\n", 
	      pname, x, minimum, maximum);
      else
	fprintf(stderr, "*  Warning  *: %s:  %f: out of range [%f... Inf]\n",
	      pname, x, minimum);
      if (x > maximum)
	{
	  fprintf(stderr, "*              taking %f instead\n",maximum);
	  return (maximum);
	}
      else
	{
	  fprintf(stderr, "*              taking %f instead\n",minimum);
	  return (minimum);
	}
    }
  return (x);
}
  
int clamp_int (int minimum, int x, int maximum, char *pname)
{
  if (x > maximum || x < minimum)
    {
      fprintf(stderr, "*  Warning  *: %s:  %d: out of range [%d...%d]\n", 
	      pname, x, minimum, maximum);
      if (x > maximum)
        {
          fprintf(stderr, "*              taking %d instead\n",maximum);
          return (maximum);
        }
      else
        {
          fprintf(stderr, "*              taking %d instead\n",minimum);
          return (minimum);
        }
    }
  return (x);
}
  
      

int read_input (char *path, int s_flag)
{

  char version[5];
  char entry_text[3 * MAXTEXTL];
  char    command[256];
  FILE *fd;
  int i, iloop, num = 0, j, k=0, totlines;
  /* int flag=0,count; */
  float infloat, f2, f3, f4, f5, f6, f7, f8;
  double dummy,sbgA;
  Vector test;
 
  fd = fopen (path, "r");
  if (!fd)
    {
      i = errno;
      if (s_flag == 1)
        fprintf(stderr, "* Warning *: fopen: %s: %s\n", path,strerror (i));
      else
	fprintf(stdout,
		"* Warning *: writing new input file %s using defaults\n", 
		path);
      /* create default file from datafile  */
      if (0 == access(datafile, R_OK))
	{
	  snprintf (command, 255, "cora_inp -s %s -i %s 2>/dev/null", 
		    datafile,path);
	  system (command);
	}
      else
	{	
	  printf("* Warning *: No read access for %s", datafile);
	  return(-1);
	}
      
      fd = fopen (path,"r");
    }
  
  totlines=0;
  while (NULL != fgets (entry_text, 256, fd))
    {
      if (entry_text[0] == '#')
        continue;
      if (entry_text[0] == ' ')
        continue;
      if (entry_text[0] == ';')
        continue;
      if (entry_text[0] == '\n')
        continue;
      if (entry_text[0] == '\t')
        continue;
      if (entry_text[0] == '*')
 	continue;
      ++totlines;
    }
  
  fclose(fd);
  fd = fopen (path, "r");
  
  j = 0;
  if (verbose > 2)
    fprintf(stderr, "Message      : listing input parameters...\n");

  while (NULL != fgets (entry_text, 128, fd))
    {
      if (verbose > 2)
	fprintf(stdout, "Message      : %s\n", entry_text);
      
      ++j;
      if (entry_text[0] == '#')
        continue;
      if (entry_text[0] == ' ')
        continue;
      if (entry_text[0] == ';')
        continue;
      if (entry_text[0] == '\n')
        continue;
      if (entry_text[0] == '\t')
        continue;
      i = 0;
      while (entry_text[i] != '\0')
        {
          if (entry_text[i] == ',')
            entry_text[i] = ' ';
          ++i;
        }
      
      i = sscanf (entry_text, "%f %f %f %f %f %f %f %f", 
		  &infloat, &f2, &f3, &f4, &f5, &f6, &f7, &f8);

      if (i < 1 && num > 1)
        {
	  fprintf (stderr, 
		   "* Warning *: Format error: %s, line %d: %s.\n", 
		   path, j - 1, entry_text);
	  return (-1);
        }
      
      /* First line is version number
       */
      if (num == 0)
	{
#ifdef HAVE_SNPRINTF
	  snprintf (version, 5, "%s", entry_text);
#else
	  sprintf  (version, "%s", entry_text);
#endif
	  if (verbose) 
	    fprintf(stdout, "Message      : Version %s\n", version);
	  
	  if(atof((char *) version) != atof((char *) VERSION))
	    {
	      fprintf(stderr,
		      "* Warning *: Wrong version of input file (%f != %f)\n",
		      atof((char *) version), atof((char *) VERSION));
	      fprintf(stderr,
		      "             setting missing parameters with defaults\n");
/*	      fclose(fd);
	      return(-1);  */
	    }
	}
      
      /* Second line is title
       */
      if (num == 1)
	{
#ifdef HAVE_SNPRINTF
	  if (0 == strncmp(entry_text,"NULL",4))
	   snprintf (fittitle, MAXTEXTL, "%s","");
	  else
	   snprintf (fittitle, MAXTEXTL, "%s", entry_text);
#else
	  if (0 == strncmp(entry_text,"NULL",4))
	   sprintf(fittitle, "%s","");
	  else
	   sprintf(fittitle, "%s", entry_text);
#endif
	  
	  k = strlen(fittitle) -1;  /* remove trailing non-sense */
	  while ( k >= 0)
	    {
	      if (fittitle[k] == '\n' ||
		  fittitle[k] == '\t' ||
		  fittitle[k] == ' ')
		{ 
		  fittitle[k] = '\0'; 
		  --k;
		}
	      else
		break;
	    }
	  if (verbose > 1) 
	    fprintf(stdout, "Message      : Title %s\n", fittitle);
	}
      
      /* Third line is number of lines
       */
      if (num == 2)
	{
	  nlines = floor (infloat);
	  if (nlines>MAXLINES)
	    {
	      printf("* Warning *: program cannot handle more than %d lines\n",
		     MAXLINES);
	      printf("          The maximum number of lines can be changed "\
		     "in cora_fit.h\n");
	      printf("          You need to change the line\n");
	      printf("          #define MAXLINES 128\n");
	      printf("          After doing that you need to type "\
		     "'make clean'\n");
	      printf("          before recompiling with make\n");
	    }
	  if(totlines != 12+nlines)
	    {
	      printf("* Warning *: invalid inputfile %s\n",path);
	      printf("             %d lines expected %d given\n",12+nlines,totlines);
	      printf("             comments have to start with # or ;\n");
	      return(-1);
	    }

	  /* allocate for nline lines 
	   */
	  vec_alloc (&llambda, nlines);
	  vec_alloc (&erlam, nlines);
	  vec_alloc (&lsigma, nlines);
	  vec_alloc (&ersig, nlines);
	  vec_alloc (&ampout, nlines);
	  vec_alloc (&eramp, nlines);
	  vec_alloc (&steplam, nlines-1);
	  vec_alloc (&stepsig, nlines-1);
	  vec_alloc (&f, nlines);

	  /* initialize line data
	   */
	  for (iloop = 0; iloop < nlines; ++iloop)
	    {
	      pro_type[iloop] = P_GAUSS;  /* line type        */
	      fit_wave[iloop] = TRUE;     /* fit wavelength ? */
	      fit_sigm[iloop] = TRUE;     /* fit sigma ?      */
	      fit_lock[iloop] = FALSE;    /* lock to previous */
	      imaxfit     = nlines;

	      ersig.vec[iloop]=-1.;
	      erlam.vec[iloop]=-1.;
	      eramp.vec[iloop]=-1.;
	    }

	  if (verbose > 1)
	    fprintf(stdout, "Message      : %d Lines :\n", nlines);
	}

      /* read the line parameters
       */
      if (num > 2 && num < nlines + 3)
        {

	  /* PORTME  read pro_type, fit_wave */
	  if (i < 2) 
	    {
	      fprintf (stderr, 
		       "* Warning *: Format error: %s, line %d: %s.\n", 
		       path, j - 1, entry_text);
	      return (-1);
	    }
      
          llambda.vec[num - 3] = infloat;
          lsigma.vec[num - 3]  = f2;

	  if (i > 2)
	    f.vec[num -3]        = f3;
	  else
	    f.vec[num -3]        = 0.0;

	  if (i > 3)
	    pro_type[num -3]     = f4;
	  else
	    pro_type[num -3]     = P_GAUSS;

	  if (i > 4)
	    fit_wave[num -3]     = f5;
	  else
	    fit_wave[num -3]     = TRUE;

	  if (i > 5)
	    fit_sigm[num -3]     = f6;
	  else
	    fit_sigm[num -3]     = TRUE;

	  if (i > 6)
	    fit_lock[num -3]     = f7;
	  else
	    fit_lock[num -3]     = FALSE;

	  if (i > 7)
	    beta[num -3]	 = f8;
	  

          if(lsigma.vec[num-3]<0.)
            {
	      fprintf(stderr, 
		      "* Warning *: lsigma:  %f: out of range [%f...]\n", 
		      lsigma.vec[num-3], 0.0);
	      lsigma.vec[num-3]=fabs(lsigma.vec[num-3]);
	      fprintf(stderr, "* Warning *: using %f instead\n", lsigma.vec[num-3]);
            }

	  if (verbose > 1) 
	    fprintf(stdout, 
		    "Message      : Line %d  Lambda %7.4f  Sigma %7.4f\n", 
		    num - 3, llambda.vec[num - 3], lsigma.vec[num - 3]);
        }


      if (num == nlines + 3)
        {
	  if (max(&lsigma) <= 0.0)
	    {
	      fprintf(stderr,
		      "** Error **: Your sigma values are all non-positive!\n"\
		      "*            You must enter at least one positive "\
		      "value.\n"\
		      "*            Non-positive values will be replaced by "\
		      "the\n"\
		      "*            largest positive value.\n");
	      return(-1);
	    }
          app_s = infloat;
          app_b = f2;
	  if (i > 2)
	    v     = f3;
	  else
	    v     = 0.0;
	  if (i > 3)
	    fixflag = f4;
	  else
	    fixflag = 1;
        }


      if (num == nlines + 4)
        {
          sbgA = infloat;
	  sbg = sbgA/*(wave.vec[wave.len-1]-wave.vec[0])/wave.len*/;
	  if(sbg<0.)
	    {
	      fprintf(stderr, "* Warning *: %s:  %f: out of "\
		      "range [%f...]\n","sbg",sbg,0.0);
	      fprintf(stderr, "             taking %f instead\n",0.0);
	      sbg=0.0;
	    }
          sbgit = floor (f2);
          bgsmooth = floor (f3);
	  sbgchi   = floor (f4);
	  medsig   = fabs(f5);
        }

      if (num == nlines + 5)
        {
          dsig    = fabs (infloat);
          sigunit = floor (f2);
          dlam    = fabs (f3);
	  ikev	  = fabs (f4);
        }

      if (num == nlines + 6)
	{
          ilamit = floor (infloat);
	  idlam = floor (f2);
	  if (f3 == 0.) f3=1.;
	  iord  = f3;
	  idsig = floor (f4);
	}

      if (num == nlines + 7)
	{
	  isigit = floor (infloat);
	}

      if (num == nlines + 8)
        if (ires!=1) ires = floor (infloat);

      if (num == nlines + 9)
	{
	  iformat = floor (infloat);
	  iportrait = floor (f2);
	}

      if (num == nlines + 10)
        {
          iplot    = floor (infloat);
          ilab     = floor (f2);
          ierrplot = floor (f3);
          ilog     = floor (f4);
          nbin     = floor (f5);
	  if (i > 5)
	   {
	    spec_only = floor (f6);
	    fit_only = floor (f7);
	   }
	  else
	   {spec_only = 0;fit_only = 0;}
        }

      if (num == nlines + 11)
        {
          nlik = floor (infloat);
          ispline = floor (f2);
        }
      ++num;
    }

  fclose(fd);
  if (verbose > 2)
    fprintf(stderr, 
	    "Message      : ...finished listing of input parameters\n");
  
  /********************************/
  /*     consistency checks       */
  /********************************/
  

  if (igraph)
    {
      if (verbose > 1)
        fprintf(stderr, "Message      : starting graphical input of lines\n");
      graph();
      if (verbose > 1)
	fprintf(stderr, "Message      : finished graphical input\n");
    }
  
  nlines    = clamp_int   (1, floor (nlines),  MAXLINES, "nlines");

  k=0;
  for(num = 0; num < nlines; ++num)
    {
      llambda.vec[num] = 
	clamp_float(min(&wave), llambda.vec[num],  max(&wave), "lambda");

      pro_type[num] =
	clamp_int  (P_MIN,      pro_type[num],     P_MAX,      "profile type");
      fit_wave[num] =
	clamp_int  (FALSE,      fit_wave[num],     TRUE,       "fit lambda");
       if (fit_wave[num] == FALSE) ++k;
      fit_sigm[num] =
	clamp_int  (FALSE,      fit_sigm[num],     TRUE,       "fit sigma");
      fit_lock[num] =
	clamp_int  (FALSE,      fit_lock[num],     TRUE,       "lock lambda");
    }
   if (k == nlines) ilamit=0;

  if (app_s <= 0.)
    {
	fprintf(stderr,"*  Warning  *: Non-positive apertures make no sense! Setting to 1\n");
	app_s=1.;
    }
  if (app_b <= 0.)
    {
        fprintf(stderr,"*  Warning  *: Non-positive apertures make no sense! Setting to 1\n");
        app_b=1.;
    }
  fixflag   = clamp_int  (0,  floor (fixflag),          1, "fixflag");
  sbg       = clamp_float(0., sbg,              FLT_MAX,  "sbg");
  sbgit     = clamp_int  (0,  floor (sbgit),          1, "sbgit");
  bgsmooth  = clamp_int  (0,  floor (bgsmooth),       1, "bgsmooth");
  sbgchi    = clamp_int  (0,  floor (sbgchi),         1, "sbgchi");
  medsig    = clamp_float (0.1, medsig,         FLT_MAX, "medsig");
  ilamit    = clamp_int  (0,  floor (ilamit),         1, "ilamit");
  isigit    = clamp_int  (0,  floor (isigit),         1, "isigit");
  spec_only = clamp_int  (0,  floor (spec_only),      1, "spec_only");
  fit_only  = clamp_int  (0,  floor (fit_only),       1, "fit_only");

  /*   prevent fitting if only plotting of spectrum is desired  */
  if (spec_only == 1)
    {
      sbgit=0;
      bgsmooth=0;
      sbgchi=0;
      ilamit=0;
      isigit=0;
      iplot=1;
      ilab=0;
      if (verbose > 1)
        fprintf(stderr, "Message      : No fit, only plotting the spectrum\n");

    }
  
  if (isigit>0)
    {
      vec_init(&test);
      vec_alloc(&test,nlines);
      j=0;
      for(i=0;i<nlines;++i)
	if (lsigma.vec[i] > 0. && fit_sigm[i] == 1) {test.vec[j]=lsigma.vec[i];++j;}
      test.len=j;  /* number of non zero sigma  */
      dsig    = clamp_float(0., fabs (dsig), min(&test),"dsig");
      vec_free(&test);
    }


  sigunit = clamp_int(0,floor (sigunit),1,"sigunit");
  if (ilamit>0)
    dlam    = clamp_float(0.,fabs (dlam),min(&llambda)-min(&wave),"dlam");
  idlam = clamp_int(0,floor (idlam),1,"idlam");
  idsig = clamp_int(0,floor (idsig),1,"idsig");
  iord  = clamp_float(0.1,iord,max(&wave)/max(&llambda),"iord");
  ires = clamp_int(0,floor (ires),1,"ires");
  iformat = clamp_int(0,floor (iformat),3,"iformat");
  iportrait = clamp_int(0,floor (iportrait),1,"iportrait");
  if(iformat>0) iformat = iformat + 3*iportrait;                        
  iplot    = clamp_int(0,floor (iplot),1,"iplot");
  ilab     = clamp_int(0,floor (ilab),1,"ilab");
  ierrplot = clamp_int(0,floor (ierrplot),1,"ierrplot");
  ilog     = clamp_int(0,floor (ilog),1,"ilog");
  nbin     = clamp_int(1,floor (nbin),20,"nbin");
  nlik = clamp_int(0,floor (nlik),1000,"nlik");
  ispline = clamp_int(0,floor (ispline),1,"ispline");
  
  
  /*  check for sufficient iteration ranges   
   */
  if(dlam<=(wave.vec[wave.len-1]-wave.vec[0])/wave.len && ilamit==1)
    {
      fprintf(stderr, "*  Warning  *: cannot iterate lambda with "\
	      "dlam=%f\n",dlam);
      ilamit=0;
    }

  /*  convert dsig to Angstroem for internal purpose
   */
  if (sigunit == 1 && lsigma.vec[0] > 10.*dlam)
  /* second condition checks if sigma is given in units A or km/s.
     assumes that sigma must be same order of magnitude as dlam. */
    {
      if (verbose) printf("Message      : converting sigmas to units A\n");
      dsig = llambda.vec[0] * dsig / C_LIGHT;
      for(i=0; i<nlines; ++i) 
	lsigma.vec[i]=llambda.vec[i]*lsigma.vec[i]/C_LIGHT;
    }
  
  /*  multiply lambdas with iord       */
  if (iord > .1) 
    for(i=0; i<nlines; ++i) 
      llambda.vec[i] = llambda.vec[i] * iord;
  
  /*   adjust non useful sigma values  */
  dummy=max(&lsigma);
  for (i=0; i<nlines; ++i) 
    if (lsigma.vec[i] <= 0.0) lsigma.vec[i]=dummy;

  /*   set locked wavelengths          */
  if (fit_lock[0] == TRUE)
    fit_lock[0] = FALSE;

  for (i=1; i < nlines; ++i) 
    if (fit_lock[i] == TRUE)
      llambda.vec[i] = llambda.vec[i-1];

  /*   define borders of sigma/lambda  */
  lsig     = min(&lsigma)-dsig;
  if (lsig < 0.) lsig = 0;
  usig     = max(&lsigma)+dsig;
  lamstart = min(&llambda) - dlam;
  lamend   = max(&llambda) + dlam;

  return(1);
}
